#![allow(dead_code)]

use std::path::{Path, PathBuf};

use regex::Regex;
use serde::{Deserialize, Serialize};

use radicle_git_ext::Oid;

use radicle::{
    git::RefString,
    identity::Did,
    node::{Alias, AliasStore, Event},
    prelude::{NodeId, RepoId},
    storage::{git::paths, ReadRepository, ReadStorage, RefUpdate},
    Profile,
};

use crate::{logger, msg};

#[derive(Debug, Eq, PartialEq, Clone, Serialize, Deserialize)]
pub enum BrokerEvent {
    Shutdown,
    BranchCreated {
        repo: RepoId,
        branch: RefString,
        tip: Oid,
        pusher: Author,
        commits: Vec<Oid>,
    },
    BranchUpdated {
        repo: RepoId,
        branch: RefString,
        tip: Oid,
        old_tip: Oid,
        pusher: Author,
        commits: Vec<Oid>,
    },
    BranchDeleted {
        repo: RepoId,
        branch: RefString,
        tip: Oid,
    },
    PatchCreated {
        repo: RepoId,
        patch: Oid,
        new_tip: Oid,
    },
    PatchUpdated {
        repo: RepoId,
        patch: Oid,
        new_tip: Oid,
    },
}

impl BrokerEvent {
    pub fn from_node_event(
        profile: &Profile,
        event: &Event,
    ) -> Result<Vec<Self>, BrokerEventError> {
        fn ref_string(s: &str) -> Result<RefString, BrokerEventError> {
            RefString::try_from(s).map_err(|e| BrokerEventError::ref_string(s.into(), e))
        }

        fn branch(ref_name: &str, update: &RefUpdate) -> Result<RefString, BrokerEventError> {
            ref_string(
                &namespaced_branch(ref_name)
                    .map_err(|_| BrokerEventError::without_namespace(ref_name, update.clone()))?,
            )
        }

        eprintln!("from_node_event: event={event:#?}");
        match event {
            Event::RefsFetched {
                remote: _,
                rid,
                updated,
            } => {
                let mut events = vec![];
                for update in updated {
                    let e = match update {
                        RefUpdate::Created { name, oid } => {
                            if let Ok(patch_id) = patch_id(name) {
                                BrokerEvent::PatchCreated {
                                    repo: *rid,
                                    patch: patch_id,
                                    new_tip: *oid,
                                }
                            } else if let Ok(branch) = namespaced_branch(name) {
                                BrokerEvent::BranchCreated {
                                    repo: *rid,
                                    branch: ref_string(&branch)?,
                                    tip: *oid,
                                    pusher: extract_author(profile, ref_string(name)?)?,
                                    commits: vec![*oid], // FIXME: OK?
                                }
                            } else {
                                logger::unhandled_ref_update(update);
                                continue;
                            }
                        }
                        RefUpdate::Updated { name, old, new } => {
                            if let Ok(patch_id) = patch_id(name) {
                                BrokerEvent::PatchUpdated {
                                    repo: *rid,
                                    patch: patch_id,
                                    new_tip: *new,
                                }
                            } else if let Ok(branch) = namespaced_branch(name) {
                                let git_repo = radicle_surf::Repository::open(paths::repository(
                                    &profile.storage,
                                    rid,
                                ))
                                .map_err(BrokerEventError::OpenRepositorySurf)?;
                                let mut commits: Vec<Oid> = git_repo
                                    .history(old)
                                    .map_err(BrokerEventError::History)?
                                    .take_while(
                                        |c| if let Ok(c) = c { c.id != *old } else { false },
                                    )
                                    .map(|r| r.map(|c| c.id))
                                    .collect::<Result<Vec<Oid>, _>>()
                                    .map_err(BrokerEventError::History)?;
                                if commits.is_empty() {
                                    commits = vec![*old];
                                }
                                BrokerEvent::BranchUpdated {
                                    repo: *rid,
                                    branch: ref_string(&branch)?,
                                    tip: *new,
                                    old_tip: *old,
                                    pusher: extract_author(profile, ref_string(name)?)?,
                                    commits,
                                }
                            } else {
                                logger::unhandled_ref_update(update);
                                continue;
                            }
                        }
                        RefUpdate::Deleted { name, oid } => BrokerEvent::BranchDeleted {
                            repo: *rid,
                            branch: branch(name, update)?,
                            tip: *oid,
                        },
                        RefUpdate::Skipped { .. } => continue,
                    };
                    logger::broker_event(&e);
                    events.push(e);
                }
                Ok(events)
            }
            Event::RefsSynced { .. }
            | Event::RefsAnnounced { .. }
            | Event::NodeAnnounced { .. }
            | Event::SeedDiscovered { .. }
            | Event::SeedDropped { .. }
            | Event::PeerConnected { .. }
            | Event::PeerDisconnected { .. }
            | Event::LocalRefsAnnounced { .. }
            | Event::UploadPack { .. }
            | Event::InventoryAnnounced { .. } => Ok(vec![]),
        }
    }

    pub fn to_request_message(&self, profile: &Profile) -> Result<msg::Request, BrokerEventError> {
        let msg = match self {
            Self::BranchCreated {
                repo,
                branch,
                tip,
                pusher,
                commits,
            } => {
                let rad_repo = profile
                    .storage
                    .repository(*repo)
                    .map_err(BrokerEventError::OpenRepository)?;
                let project_info = rad_repo
                    .project()
                    .map_err(BrokerEventError::OpenRepository)?;
                let repository = msg::Repository {
                    id: *repo,
                    name: project_info.name().to_string(),
                    description: project_info.description().to_string(),
                    private: !rad_repo
                        .identity()
                        .map_err(BrokerEventError::GetIdentity)?
                        .visibility
                        .is_public(),
                    default_branch: project_info.default_branch().to_string(),
                    delegates: rad_repo
                        .delegates()
                        .map_err(BrokerEventError::GetDelegates)?
                        .iter()
                        .copied()
                        .collect(),
                };

                msg::Request::Trigger {
                    common: msg::EventCommonFields {
                        version: msg::PROTOCOL_VERSION,
                        event_type: msg::EventType::Push,
                        repository,
                    },
                    patch: None,
                    push: Some(msg::PushEvent {
                        pusher: msg::Author {
                            id: pusher.id,
                            alias: pusher.alias.clone(),
                        },
                        before: *tip, // FIXME: is this right?
                        after: *tip,
                        branch: branch.to_string(),
                        commits: commits.clone(),
                    }),
                }
            }
            Self::BranchUpdated {
                repo,
                branch,
                tip,
                old_tip,
                pusher,
                commits,
            } => {
                let rad_repo = profile
                    .storage
                    .repository(*repo)
                    .map_err(BrokerEventError::OpenRepository)?;
                let project_info = rad_repo
                    .project()
                    .map_err(BrokerEventError::OpenRepository)?;
                let repository = msg::Repository {
                    id: *repo,
                    name: project_info.name().to_string(),
                    description: project_info.description().to_string(),
                    private: !rad_repo
                        .identity()
                        .map_err(BrokerEventError::GetIdentity)?
                        .visibility
                        .is_public(),
                    default_branch: project_info.default_branch().to_string(),
                    delegates: rad_repo
                        .delegates()
                        .map_err(BrokerEventError::GetDelegates)?
                        .iter()
                        .copied()
                        .collect(),
                };

                msg::Request::Trigger {
                    common: msg::EventCommonFields {
                        version: msg::PROTOCOL_VERSION,
                        event_type: msg::EventType::Push,
                        repository,
                    },
                    patch: None,
                    push: Some(msg::PushEvent {
                        pusher: msg::Author {
                            id: pusher.id,
                            alias: pusher.alias.clone(),
                        },
                        before: *old_tip,
                        after: *tip,
                        branch: branch.to_string(),
                        commits: commits.clone(),
                    }),
                }
            }
            Self::BranchDeleted { .. } => return Err(BrokerEventError::NoDeletionTrigger),
            _ => unreachable!(),
        };
        Ok(msg)
    }
}

// Fields describing the author of a change.
#[derive(Debug, Clone, Eq, PartialEq, Serialize, Deserialize)]
pub struct Author {
    pub id: Did,
    pub alias: Option<Alias>,
}

impl std::fmt::Display for Author {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        write!(f, "{}", self.id)?;
        if let Some(alias) = &self.alias {
            write!(f, " ({})", alias)?;
        }
        Ok(())
    }
}

fn did_to_author(profile: &Profile, did: &Did) -> Result<Author, BrokerEventError> {
    let alias = profile.aliases().alias(did);
    Ok(Author { id: *did, alias })
}

fn parse_nid_from_refstring(name: &RefString) -> Result<Option<NodeId>, BrokerEventError> {
    const PAT: &str = r"^refs/namespaces/(?P<nid>[^/]+)/";
    let pat = Regex::new(PAT).map_err(|e| BrokerEventError::Regex(PAT, e))?;
    if let Some(captures) = pat.captures(name.as_str()) {
        if let Some(m) = captures.name("nid") {
            if let Ok(parsed) = m.as_str().parse() {
                return Ok(Some(parsed));
            }
        }
    }
    Ok(None)
}

fn extract_author(profile: &Profile, ref_name: RefString) -> Result<Author, BrokerEventError> {
    if let Some(nid) = parse_nid_from_refstring(&ref_name)? {
        did_to_author(profile, &Did::from(nid))
    } else {
        Err(BrokerEventError::NoNid(ref_name.clone()))
    }
}

pub struct BrokerEvents {
    events: Vec<BrokerEvent>,
}

impl BrokerEvents {
    pub fn from_file(filename: &Path) -> Result<Self, BrokerEventError> {
        let events =
            std::fs::read(filename).map_err(|e| BrokerEventError::read_file(filename, e))?;
        let events =
            String::from_utf8(events).map_err(|e| BrokerEventError::not_utf8(filename, e))?;
        let events: Result<Vec<BrokerEvent>, _> =
            events.lines().map(serde_json::from_str).collect();
        let events = events.map_err(|e| BrokerEventError::not_json(filename, e))?;

        Ok(Self { events })
    }

    pub fn iter(&self) -> impl Iterator<Item = &BrokerEvent> {
        self.events.iter()
    }
}

#[derive(Debug, thiserror::Error)]
pub enum BrokerEventError {
    /// Regex compilation error.
    #[error("programming error in regular expression {0:?}")]
    Regex(&'static str, regex::Error),

    /// Can't find node id in ref name.
    #[error("ref name does not include node id")]
    NoNid(RefString),

    #[error("updated ref name has no name space: {0:?}): from {1:#?}")]
    WithoutNamespace(String, RefUpdate),

    #[error("failed to create a RefString from {0:?}")]
    RefString(String, radicle::git::fmt::Error),

    #[error("failed to read broker events file {0}")]
    ReadFile(PathBuf, #[source] std::io::Error),

    #[error("broker events file is not UTF8: {0}")]
    NotUtf8(PathBuf, #[source] std::string::FromUtf8Error),

    #[error("broker events file is not valid JSON: {0}")]
    NotJson(PathBuf, #[source] serde_json::Error),

    #[error("failed to open Git repository in node storage")]
    OpenRepository(#[source] radicle::storage::RepositoryError),

    #[error("failed to open Git repository in node storage")]
    OpenRepositorySurf(#[source] radicle_surf::Error),

    #[error("failed to read repository identity")]
    GetIdentity(#[source] radicle::storage::RepositoryError),

    #[error("failed to read repository delegates")]
    GetDelegates(#[source] radicle::storage::RepositoryError),

    /// Can't get history from Git.
    #[error("failed to get history from Git")]
    History(#[source] radicle_surf::Error),

    #[error("programming error: can't create a trigger request message from branch deletion")]
    NoDeletionTrigger,
}

impl BrokerEventError {
    fn without_namespace(refname: &str, update: RefUpdate) -> Self {
        Self::WithoutNamespace(refname.into(), update)
    }

    fn ref_string(name: String, err: radicle::git::fmt::Error) -> Self {
        Self::RefString(name, err)
    }

    fn read_file(filename: &Path, err: std::io::Error) -> Self {
        Self::ReadFile(filename.into(), err)
    }

    fn not_utf8(filename: &Path, err: std::string::FromUtf8Error) -> Self {
        Self::NotUtf8(filename.into(), err)
    }

    fn not_json(filename: &Path, err: serde_json::Error) -> Self {
        Self::NotJson(filename.into(), err)
    }
}

#[cfg(test)]
mod test {
    use super::*;
    use crate::test::MockNode;
    use radicle::{prelude::NodeId, storage::RefUpdate};
    use std::str::FromStr;

    const MAIN_BRANCH_REF_NAME: &str =
        "refs/namespaces/z6MkgEMYod7Hxfy9qCvDv5hYHkZ4ciWmLFgfvm3Wn1b2w2FV/refs/heads/main";
    const MAIN_BRANCH_NAME: &str = "main";
    const MAIN_BRANCH_NID: &str = "z6MkgEMYod7Hxfy9qCvDv5hYHkZ4ciWmLFgfvm3Wn1b2w2FV";

    const PATCH_REF_NAME: &str =
        "refs/namespaces/NID/refs/heads/patches/f9fa90725474de9002be503ae3cda4670c9a174";
    const PATCH_ID: &str = "f9fa90725474de9002be503ae3cda4670c9a174";

    fn nid() -> NodeId {
        const NID: &str = "z6MkgEMYod7Hxfy9qCvDv5hYHkZ4ciWmLFgfvm3Wn1b2w2FV";
        NodeId::from_str(NID).unwrap()
    }

    fn rid() -> RepoId {
        const RID: &str = "rad:z3gqcJUoA1n9HaHKufZs5FCSGazv5";
        RepoId::from_urn(RID).unwrap()
    }

    fn oid_from(oid: &str) -> Oid {
        Oid::try_from(oid).unwrap()
    }

    fn oid() -> Oid {
        const OID: &str = "ff3099ba5de28d954c41d0b5a84316f943794ea4";
        oid_from(OID)
    }

    fn refstring(s: &str) -> RefString {
        RefString::try_from(s).unwrap()
    }

    #[test]
    fn nothing_updated() {
        let event = Event::RefsFetched {
            remote: nid(),
            rid: rid(),
            updated: vec![],
        };
        let profile = MockNode::new().unwrap().profile().unwrap();
        let result = BrokerEvent::from_node_event(&profile, &event);
        assert!(result.is_ok());
        assert_eq!(result.unwrap(), vec![]);
    }

    #[test]
    fn skipped() {
        let event = Event::RefsFetched {
            remote: nid(),
            rid: rid(),
            updated: vec![RefUpdate::Skipped {
                name: refstring(MAIN_BRANCH_REF_NAME),
                oid: oid(),
            }],
        };

        let profile = MockNode::new().unwrap().profile().unwrap();
        let result = BrokerEvent::from_node_event(&profile, &event);
        assert!(result.is_ok());
        assert_eq!(result.unwrap(), vec![]);
    }

    #[test]
    fn branch_created() {
        let rid = rid();
        let main = refstring(MAIN_BRANCH_NAME);
        let oid = oid();
        let event = Event::RefsFetched {
            remote: nid(),
            rid,
            updated: vec![RefUpdate::Created {
                name: refstring(MAIN_BRANCH_REF_NAME),
                oid,
            }],
        };
        let profile = MockNode::new().unwrap().profile().unwrap();
        let x = BrokerEvent::from_node_event(&profile, &event);
        eprintln!("result: {x:#?}");
        match x {
            Err(_) => panic!("should succeed"),
            Ok(events) if !events.is_empty() => {
                for e in events {
                    match e {
                        BrokerEvent::BranchCreated {
                            repo,
                            branch,
                            tip,
                            pusher: _,
                            commits: _,
                        } if repo == rid && branch == main && tip == oid => {}
                        _ => panic!("should not succeed that way"),
                    }
                }
            }
            Ok(_) => panic!("empty list of events should not happen"),
        }
    }

    // #[test]
    // fn branch_updated() {
    //     let rid = rid();
    //     let main = refstring(MAIN_BRANCH_NAME);
    //     let oid = oid();
    //     let event = Event::RefsFetched {
    //         remote: nid(),
    //         rid,
    //         updated: vec![RefUpdate::Updated {
    //             name: refstring(MAIN_BRANCH_REF_NAME),
    //             old: oid,
    //             new: oid,
    //         }],
    //     };
    //     let profile = MockNode::new().unwrap().profile().unwrap();
    //     let x = BrokerEvent::from_node_event(&profile, &event);
    //     eprintln!("result: {x:#?}");
    //     match x {
    //         Err(_) => panic!("should succeed"),
    //         Ok(events) if !events.is_empty() => {
    //             for e in events {
    //                 match e {
    //                     BrokerEvent::BranchUpdated {
    //                         repo,
    //                         branch,
    //                         tip,
    //                         old_tip,
    //                         pusher: _,
    //                         commits: _,
    //                     } if repo == rid && branch == main && tip == oid && old_tip == oid => {}
    //                     _ => panic!("should not succeed that way"),
    //                 }
    //             }
    //         }
    //         Ok(_) => panic!("empty list of events should not happen"),
    //     }
    // }

    #[test]
    fn branch_deleted() {
        let rid = rid();
        let main = refstring(MAIN_BRANCH_NAME);
        let oid = oid();
        let event = Event::RefsFetched {
            remote: nid(),
            rid,
            updated: vec![RefUpdate::Deleted {
                name: refstring(MAIN_BRANCH_REF_NAME),
                oid,
            }],
        };
        let profile = MockNode::new().unwrap().profile().unwrap();
        let x = BrokerEvent::from_node_event(&profile, &event);
        eprintln!("result: {x:#?}");
        match x {
            Err(_) => panic!("should succeed"),
            Ok(events) if !events.is_empty() => {
                for e in events {
                    match e {
                        BrokerEvent::BranchDeleted { repo, branch, tip }
                            if repo == rid && branch == main && tip == oid => {}
                        _ => panic!("should not succeed that way"),
                    }
                }
            }
            Ok(_) => panic!("empty list of events should not happen"),
        }
    }

    #[test]
    fn patch_created() {
        let rid = rid();
        let patch_id = oid_from(PATCH_ID);
        let oid = oid();
        let event = Event::RefsFetched {
            remote: nid(),
            rid,
            updated: vec![RefUpdate::Created {
                name: refstring(PATCH_REF_NAME),
                oid,
            }],
        };
        let profile = MockNode::new().unwrap().profile().unwrap();
        let x = BrokerEvent::from_node_event(&profile, &event);
        eprintln!("result: {x:#?}");
        match x {
            Err(_) => panic!("should succeed"),
            Ok(events) if !events.is_empty() => {
                for e in events {
                    match e {
                        BrokerEvent::PatchCreated {
                            repo,
                            patch,
                            new_tip,
                        } if repo == rid && patch == patch_id && new_tip == oid => {}
                        _ => panic!("should not succeed that way"),
                    }
                }
            }
            Ok(_) => panic!("empty list of events should not happen"),
        }
    }

    #[test]
    fn patch_updated() {
        let rid = rid();
        let patch_id = oid_from(PATCH_ID);
        let oid = oid();
        let event = Event::RefsFetched {
            remote: nid(),
            rid,
            updated: vec![RefUpdate::Updated {
                name: refstring(PATCH_REF_NAME),
                old: oid,
                new: oid,
            }],
        };
        let profile = MockNode::new().unwrap().profile().unwrap();
        let x = BrokerEvent::from_node_event(&profile, &event);
        eprintln!("result: {x:#?}");
        match x {
            Err(_) => panic!("should succeed"),
            Ok(events) if !events.is_empty() => {
                for e in events {
                    match e {
                        BrokerEvent::PatchUpdated {
                            repo,
                            patch,
                            new_tip,
                        } if repo == rid && patch == patch_id && new_tip == oid => {}
                        _ => panic!("should not succeed that way"),
                    }
                }
            }
            Ok(_) => panic!("empty list of events should not happen"),
        }
    }
}

fn namespaced_branch(refname: &str) -> Result<String, ParseError> {
    const PAT_BRANCH: &str = r"^refs/namespaces/[^/]+/refs/heads/(.+)$";
    let push_re = Regex::new(PAT_BRANCH).map_err(|e| ParseError::Regex(PAT_BRANCH, e))?;
    if let Some(push_captures) = push_re.captures(refname) {
        if let Some(branch) = push_captures.get(1) {
            return Ok(branch.as_str().to_string());
        }
    }
    Err(ParseError::NotBranch(refname.into()))
}

fn patch_id(refname: &str) -> Result<Oid, ParseError> {
    eprintln!("refname: {refname:?}");
    const PAT_PATCH: &str = r"^refs/namespaces/[^/]+/refs/heads/patches/([^/]+)$";
    let patch_re = Regex::new(PAT_PATCH).map_err(|e| ParseError::regex(PAT_PATCH, e))?;
    if let Some(patch_captures) = patch_re.captures(refname) {
        eprintln!("captures: {patch_captures:?}");
        if let Some(patch_id) = patch_captures.get(1) {
            eprintln!("patch_id: {patch_id:?}");
            let oid = Oid::try_from(patch_id.as_str())
                .map_err(|e| ParseError::oid(patch_id.as_str(), e))?;
            eprintln!("oid: {oid:?}");
            return Ok(oid);
        }
    }

    Err(ParseError::not_patch(refname))
}

#[derive(Debug, thiserror::Error)]
enum ParseError {
    #[error("programming error: unacceptable regular expression {0:?}")]
    Regex(&'static str, regex::Error),

    #[error("Git ref name without name space: {0:?}")]
    NotBranch(String),

    #[error("unacceptable Git ref for patch: {0:?}")]
    NotPatch(String),

    #[error("Git ref name includes unacceptable Git object id: {0:?}")]
    Oid(String, radicle::git::raw::Error),
}

impl ParseError {
    fn regex(pattern: &'static str, err: regex::Error) -> Self {
        Self::Regex(pattern, err)
    }

    fn not_branch(refname: &str) -> Self {
        Self::NotBranch(refname.into())
    }

    fn not_patch(refname: &str) -> Self {
        Self::NotPatch(refname.into())
    }

    fn oid(refname: &str, err: radicle::git::raw::Error) -> Self {
        Self::Oid(refname.into(), err)
    }
}

#[cfg(test)]
mod test_namespaced_branch {
    use super::{namespaced_branch, ParseError};

    #[test]
    fn empty() {
        assert!(matches!(
            namespaced_branch(""),
            Err(ParseError::NotBranch(_))
        ));
    }

    #[test]
    fn lacks_namespace() {
        assert!(matches!(
            namespaced_branch(""),
            Err(ParseError::NotBranch(_))
        ));
    }

    #[test]
    fn has_namespace() {
        let x = namespaced_branch("refs/namespaces/DID/refs/heads/main");
        assert!(x.is_ok());
        assert_eq!(x.unwrap(), "main");
    }

    #[test]
    fn has_namespace_with_path() {
        let x = namespaced_branch("refs/namespaces/DID/refs/heads/liw/debug/this/path");
        assert!(x.is_ok());
        assert_eq!(x.unwrap(), "liw/debug/this/path");
    }
}

#[cfg(test)]
mod test_patch_id {
    use super::{patch_id, Oid, ParseError};

    #[test]
    fn empty() {
        assert!(matches!(patch_id(""), Err(ParseError::NotPatch(_))));
    }

    #[test]
    fn lacks_namespace() {
        assert!(matches!(patch_id(""), Err(ParseError::NotPatch(_))));
    }

    #[test]
    fn has_namespace() {
        let x = patch_id(
            "refs/namespaces/DID/refs/heads/patches/f9fa90725474de9002be503ae3cda4670c9a1741",
        );
        assert!(x.is_ok());
        assert_eq!(
            x.unwrap(),
            Oid::try_from("f9fa90725474de9002be503ae3cda4670c9a1741").unwrap()
        );
    }

    #[test]
    fn has_namespace_with_path() {
        assert!(matches!(
            patch_id("refs/namespaces/DID/refs/heads/patches/coffee/beef"),
            Err(ParseError::NotPatch(_))
        ));
    }
}
